import errno
import os
import signal
import time

import billiard
from celery.bootsteps import StartStopStep
from celery.contrib.abortable import AbortableAsyncResult, AbortableTask
from celery.utils.log import get_task_logger
from celery_singleton import Singleton

__all__ = ('EternalTask', 'EternalProcessTask')

# Logging
log = get_task_logger(__name__)


class AbortStep(StartStopStep):
    """Boot step to abort all tasks on shutdown."""
    requires = {'celery.worker.components:Pool'}

    def stop(self, worker):
        for request in worker.state.active_requests:
            AbortableAsyncResult(request.task_id).abort()


class EternalTask(AbortableTask, Singleton):
    """Base class for a task that should run forever, and should be restarted
    if it ever exits. The task should periodically check
    :meth:`~celery.contrib.abortable.AbortableTask.is_aborted`
    and exit gracefully if it is set. During a warm shutdown, we will attempt
    to abort the task.

    Example
    -------

    To create an abortable task, call the :meth:`~celery.Celery.task` decorator
    with the keyword argument ``base=EternalTask``.

    Generally, you should also pass ``bind=True`` so that your task function
    has access to the :class:`~celery.app.task.Task` instance and can call
    ``self.is_aborted()``.

    Here is an example eternal task that calls a fictional function
    `do_some_work()` in a loop::

        @app.task(base=EternalTask, bind=True, ignore_result=True)
        def long_running_task(self):
            while not self.is_aborted():
                do_some_work()

    """

    ignore_result = True

    @classmethod
    def on_bound(cls, app):
        app.add_periodic_task(1.0, cls)
        app.steps['worker'].add(AbortStep)

    def __exit_message(self):
        if self.is_aborted():
            log.info('Eternal process exited early due to abort')
        else:
            log.error('Eternal process exited early!')

    def on_failure(self, *args, **kwargs):
        super(EternalTask, self).on_failure(*args, **kwargs)
        self.__exit_message()

    def on_success(self, *args, **kwargs):
        super(EternalTask, self).on_success(*args, **kwargs)
        self.__exit_message()


class EternalProcessTask(EternalTask):
    """Base class for an eternal task that runs in a subprocess.

    This is useful for tasks that do not check the method
    :meth:`~celery.contrib.abortable.AbortableTask.is_aborted` but can be
    stopped by a :obj:`KeyboardInterrupt` triggered by receving the signal
    :data:`SIGINT`.

    The task itself launches and supervises a subprocess that runs the
    function. The subprocess has Python's default :data:`SIGINT` handler
    installed.
    """

    def __call__(self, *args, **kwargs):
        process = billiard.Process(target=self.run, *args, **kwargs)
        old_handler = signal.signal(signal.SIGINT, signal.default_int_handler)
        try:
            process.start()
        finally:
            signal.signal(signal.SIGINT, old_handler)
        while process.is_alive() and not self.is_aborted():
            time.sleep(1)
        if self.is_aborted():
            try:
                os.kill(process.pid, signal.SIGINT)
            except OSError as e:
                if e.errno != errno.ESRCH:
                    log.exception('Failed to kill subprocess')
        process.join(1)
        process.terminate()



from __future__ import absolute_import
from multiprocessing import Process
import os
import signal
from time import sleep

from celery import Celery
from celery.signals import worker_process_shutdown
from kombu.exceptions import OperationalError
import pytest

from .. import EternalTask, EternalProcessTask

# Celery application object.
# Use redis backend, because it supports locks (and thus singleton tasks).
app = Celery(__name__, broker='redis://')
app.conf['result_backend'] = app.conf.broker_url

# Only run these tests if a Redis server is running.
try:
    app.connection().ensure_connection(max_retries=1)
except OperationalError:
    pytestmark = pytest.mark.skip('No Redis server is running.')


def touch(path):
    """Touch a file."""
    with open(path, 'w'):
        pass


@app.task(base=EternalTask, bind=True, shared=False)
def example_task_aborts_gracefully(self):
    while not self.is_aborted():
        sleep(0.1)
        touch(os.path.join(os.environ['COV_TMP'], 'aborts_gracefully'))


@app.task(base=EternalTask, shared=False)
def example_task_always_succeeds():
    sleep(0.1)
    touch(os.path.join(os.environ['COV_TMP'], 'always_succeeds'))


@app.task(base=EternalTask, shared=False)
def example_task_always_fails():
    sleep(0.1)
    touch(os.path.join(os.environ['COV_TMP'], 'always_fails'))
    raise RuntimeError('Expected to fail!')


@app.task(base=EternalProcessTask, shared=False)
def example_task_never_returns():
    while True:
        touch(os.path.join(os.environ['COV_TMP'], 'never_returns'))
        sleep(1)


@app.task(base=EternalProcessTask, shared=False)
def example_task_ignores_sigint():
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    while True:
        touch(os.path.join(os.environ['COV_TMP'], 'ignores_sigint'))
        sleep(1)


# Only needed if we are measuring test coverage
try:
    from pytest_cov.embed import multiprocessing_finish
except ImportError:
    pass
else:
    @worker_process_shutdown.connect
    def on_worker_process_shutdown(*args, **kwargs):
        multiprocessing_finish()


@pytest.fixture
def start_test_app_worker(tmpdir):
    """Start up a worker for the test app."""
    os.environ['COV_TMP'] = str(tmpdir)
    argv = ['worker', '-B', '-c', '5', '-l', 'info',
            '-s', str(tmpdir / 'celerybeat-schedule')]
    p = Process(target=app.worker_main, args=(argv,))
    p.daemon = True
    p.start()
    yield
    p.terminate()
    p.join()
    app.control.purge()
    del os.environ['COV_TMP']


def test_eternal(start_test_app_worker, tmpdir):
    """Test worker with two eternal tasks: one that always succeeds,
    and one that always fails."""
    filenames = ['aborts_gracefully',
                 'always_succeeds',
                 'always_fails',
                 'never_returns',
                 'ignores_sigint']
    for i in range(100):
        finished = all(os.path.exists(str(tmpdir / _)) for _ in filenames)
        if finished:
            break
        sleep(0.1)
    assert finished
